unit OctTreeClass;
//*******************************************************
// Octtree class.
//*******************************************************

interface

uses Windows, Graphics, sysutils, Classes, contnrs;

type
  TNorthOrSouth = (North, South);
  TEastOrWest = (East, West);
  TUpOrDown = (Up, Down);

  TOPoint = record
    X: double;
    Y: double;
    Z: double;
    Data : TList;
  end;
  POPoint = ^TOPoint;

  TPointerArray = array of Pointer;

  TOctPoint =  record
    X: double;
    Y: double;
    Z: double;
    Distance : double;
    Data : TPointerArray;
  end;
//  TOctPoint records contain the X, Y, and Z coordinates of a location in a
//  TRbwOctTree, the distance from that point to a location for which a
//  search was performed, and an array of pointers to the data associated
//  with the location.

  TOctPointArray = array of TOctPoint;

  TOctPointInRegion =  record
    X: double;
    Y: double;
    Z: double;
    Data : TPointerArray;
  end;
//  TOctPointInRegion records contain the X, Y, and Z coordinates of a location
//  in a TRbwOctTree, and an array of pointers to the data associated
//  with the location. These points are all located inside the extent of a
//  T3DBlock (see below) that was used in a search.

  TOctPointInRegionArray = array of TOctPointInRegion;

  TExpandDirection3 = (ed3North, ed3South, ed3East, ed3West, ed3Up, ed3Down);

  T3DBlock = record
    XMin : double;
    XMax : double;
    YMin : double;
    YMax : double;
    ZMin : double;
    ZMax : double;
  end;
//  T3DBlock delineates a region of space. T3DBlock's are used to perform
//  a search for locations inside a TRbwOctTree.

  EOTreeError = Class(Exception);
//  EOTreeError is a class for exceptions raised by TRbwOctTree.

  TRbwOctTree = class;

  TOtreeNode = class(TObject)
  // TOtreeNode is the most of the real work of a TRbwOctTree.
  private
    FNumPts: Integer;
    procedure ResetCount;
    procedure SetNumPts(const Value: Integer);
    function GetPoints(Index: Integer): TOPoint;
    procedure FindNearestPoints(const CenterX, CenterY, CenterZ: double;
      const Count: Integer; const List: TObjectList);
  protected
    Children               : array [TNorthOrSouth, TEastOrWest, TUpOrDown] of TOtreeNode;
    XMin, XMax, YMin, YMax, ZMin, ZMax : double;
    Pts                    : array of TOPoint;
    Parent : TOtreeNode;
    Tree : TRbwOctTree;
    property NumPts  : Integer read FNumPts write SetNumPts;
    procedure AddPoint(X, Y, Z : double; Data : pointer);
    procedure Clear;
    procedure CheckNearbyLeaves(exclude : TOtreeNode;
      var best_leaf : TOtreeNode; X, Y, Z : double; var best_i : integer;
      var best_dist2, best_dist : double);
    constructor Create(ATree: TRbwOctTree); overload;
    constructor Create(const x_min, x_max, y_min, y_max, z_min, z_max : double;
      const ParentNode : TOtreeNode; ATree : TRbwOctTree); overload;
    procedure ExpandBounds(XYZ : double; ExpandDirection : TExpandDirection3);
    function FindClosestPoint(const X, Y, Z: double): TOPoint;
    // FindClosestPoint returns the X, Y, Z locations and first piece of data in the
    // point closest to the value specified for X, Y, and Z.
    procedure FindPoint(var X, Y, Z : double; var Data : pointer);
    procedure FindClosestPointsData(var X, Y, Z : double; var Data : TPointerArray);
    function LocateLeaf(X, Y, Z : double; Siblings : TStack) : TOtreeNode;
    procedure NearPointInLeaf(X, Y, Z : double; var best_i : integer;
      var best_dist2 : double);
    function Xmid : double;
    function Ymid : double;
    function Zmid : double;
    property Points[Index: Integer]: TOPoint read GetPoints;
    procedure RemovePoint(X, Y, Z: double; Data: pointer);
    procedure FindPointsInSphere(Const CenterX, CenterY, CenterZ, Radius,
      RadiusSquared : double; const List : TList);
    procedure FindPointsInBlock(const Block: T3DBlock; const List: TList);
  public
    destructor Destroy; override;
  end;

  TRbwOctTree = class(TComponent)
  private
    FOTreeNode : TOtreeNode;
    FMaxPoints: integer;
    function GetXMax : double;
    function GetXMin : double;
    function GetYMax : double;
    function GetYMin : double;
    function GetZMax : double;
    function GetZMin : double;
    procedure SetXMax(const AValue : double);
    procedure SetXMin(const AValue : double);
    procedure SetYMax(const AValue : double);
    procedure SetYMin(const AValue : double);
    procedure SetZMax(const AValue : double);
    procedure SetZMin(const AValue : double);
    function GetCount : integer;
    function GetPoints(Index: Integer): TOctPoint;
    procedure SetMaxPoints(const Value: integer);
  public
    procedure AddPoint(X, Y, Z : double; Data : pointer);
    // AddPoint stores Data at location X, Y, Z.
    //
    // If there is already a point at that location the number of pointers
    // stored at that point will be increased by one and the new data will be
    // added at the end of the list of data for the point. Count will be
    // unaffected.
    //
    // If there is not already a point at that location, one will be created
    // and the data will be associated with it, and the Count will be
    // incremented by one.
    //
    // If required, Xmax, XMin, Ymax, Ymin, Zmax or Zmin will be adjusted to
    // accomodate the new data point.
    //
    // The Data associated with a location is not owned by the TRbwOctTree;
    // users are responsible for freeing that memory when it is no longer
    // needed.
    Procedure Clear;
    // Clear deletes all data locations in the TRbwOctTree and resets the
    // Count to 0.
    //
    // The Data associated with a location is not owned by the TRbwOctTree;
    // users are responsible for freeing that memory when it is no longer
    // needed.
    property Count : integer read GetCount;
    // Count is the number of unique locations stored in a TRbwOctTree.
    constructor Create(AOwner: TComponent); override;
    // Creates an instance of TRbwOctTree;
    destructor Destroy; override;
    // Destroys and instance of TRbwOctTree. Applications should not call
    // Destroy directly. Instead call Free;
    procedure FindClosestPointsData(var X, Y, Z : double; var Data : TPointerArray);
    // FindClosestPointsData locates the location nearest to the X, Y, Z values
    // provided as input.  It than returns the X, Y and Z values of that
    // location. The length of Data is changed to the number of pointers stored
    // at the location that was found and the Data pointers are copied into
    // Data.
    procedure FindNearestPoints(const CenterX, CenterY, CenterZ: double;
      const Count: Integer; var Points : TOctPointArray);
    // FindNearestPoints finds the Count locations in the TRbwOctTree that are
    // closest to CenterX, CenterY, CenterZ and returns them and their
    // associated data in Points. In the event of ties, the length of Points
    // may be larger than Count.  If the number of locations in the TRbwOctTree
    // is less than Count, all the locations will be returned.  Points is
    // sorted from the closest locations to CenterX, CenterY, CenterZ to those
    // that are most distant.
    //
    // See also FirstNearestPoint, NearestPointsFirstData
    procedure FindPointsInBlock(const Block: T3DBlock;
      var Points: TOctPointInRegionArray);
    // FindPointsInBlock finds all locations stored in the TRbwOctTree that
    // are inside Block and returns them in Points along with their associated
    // data. Points is not sorted.
    procedure FindPointsInSphere(Const CenterX, CenterY, CenterZ, Radius : double;
      var Points : TOctPointInRegionArray);
    // FindPointsInSphere finds all the locations stored in the TRbwOctTree
    // that are within a distance of Radius from CenterX, CenterY, CenterZ and
    // returns them and their associated data in Points.  Points is not sorted.
    procedure FirstNearestPoint(var X, Y, Z : double; var Data : pointer);
    // FirstNearestPoint, finds the location in the TRbwOctTree that is
    // closest to X, Y, Z and returns it in X, Y, Z.  The associated data is
    // returned in Data. In the event of ties in location, the location that
    // is returned is chosen arbitrarily from the locations that are closest.
    // If more than one piece of data is stored at X, Y, Z, the one that is
    // returned is chosen arbitrarily.
    //
    // See also: FindNearestPoints, NearestPointsFirstData
    function NearestPointsFirstData(X, Y, Z : double) : pointer;
    // NearestPointsFirstData finds the location stored in the TRbwOctTree
    // that is closest to X, Y, Z and returns the first piece of data
    // associated with that location.
    //
    // See also: FindNearestPoints, FirstNearestPoint
    property Points[Index: Integer]: TOctPoint read GetPoints;
    // Points can be treated as an indexed array of the locations stored
    // in the TRbwOctTree.
    procedure RemovePoint(X, Y, Z : double; Data : pointer);
    // RemovePoint removes Data at location X, Y, Z.
    //
    // If the number of data pointers stored at X, Y, Z is greater than 1, Data
    // will be deleted from the list of data pointers and Count will be
    // unaffected.
    //
    // If there is only one data pointer stored at X, Y, Z, that location will
    // be deleted and the Count will be decremented by one.
    //
    // The Data associated with a location is not owned by the TRbwOctTree;
    // users are responsible for freeing that memory when it is no longer
    // needed.
  published
//    property DestroyProcedure : TDestroyDataProcedure read FDestroyProcedure write SetDestroyProcedure;
    // Setting Xmax, XMin, Ymax, and Ymin can increase the bounds specified in Create but not decrease them.
    property MaxPoints : integer read FMaxPoints write SetMaxPoints;
    // MaxPoints is the maximum number of data locations in any node of a
    // TRbwOctTree. MaxPoints affects the performance of the TRbwOctTree
    // in the retrieval of data in a manner that depends on the particular
    // application.
    Property XMax : double read GetXMax write SetXMax;
    // Xmax is the largest X value of any of the data points that are expected
    // to be or have already been added to the TRbwOctTree.  Although it is
    // not required that Xmax be greater than the X coordinate of all locations
    // added to the TRbwOctTree, performance may suffer if too many of them
    // are greater than Xmax.
    Property XMin : double read GetXMin write SetXMin;
    // XMin is the smallest X value of any of the data points that are expected
    // to be or have already been added to the TRbwOctTree.  Although it is
    // not required that XMin be less than the X coordinate of all locations
    // added to the TRbwOctTree, performance may suffer if too many of them
    // are less than XMin.
    Property YMax : double read GetYMax write SetYMax;
    // Ymax is the largest Y value of any of the data points that are expected
    // to be or have already been added to the TRbwOctTree.  Although it is
    // not required that Ymax be greater than the Y coordinate of all locations
    // added to the TRbwOctTree, performance may suffer if too many of them
    // are greater than Ymax.
    Property YMin : double read GetYMin write SetYMin;
    // Ymin is the smallest Y value of any of the data points that are expected
    // to be or have already been added to the TRbwOctTree.  Although it is
    // not required that Ymin be less than the Y coordinate of all locations
    // added to the TRbwOctTree, performance may suffer if too many of them
    // are less than Ymin.
    Property ZMax : double read GetZMax write SetZMax;
    // ZMax is the largest Z value of any of the data points that are expected
    // to be or have already been added to the TRbwOctTree.  Although it is
    // not required that ZMax be greater than the Z coordinate of all locations
    // added to the TRbwOctTree, performance may suffer if too many of them
    // are greater than ZMax.
    Property ZMin : double read GetZMin write SetZMin;
    // ZMin is the smallest Z value of any of the data points that are expected
    // to be or have already been added to the TRbwOctTree.  Although it is
    // not required that ZMin be less than the Z coordinate of all locations
    // added to the TRbwOctTree, performance may suffer if too many of them
    // are less than ZMin.
  end;

procedure Register;
  
implementation

type
  TSelectNode = class(TObject)
    Distance : double;
    Point : TOPoint;
  end;

function SelectNodeCompare(Item1, Item2: Pointer): Integer;
var
  Node1, Node2 : TSelectNode;
  Delta : double;
begin
  Node1 := Item1;
  Node2 := Item2;
  Delta := Node2.Distance - Node1.Distance;
  if Delta > 0 then
  begin
    result := -1;
  end
  else if Delta < 0 then
  begin
    result := 1;
  end
  else
  begin
    result := 0;
  end;

end;

// Create the node.
constructor TOtreeNode.Create(ATree : TRbwOctTree);
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
  ud : TUpOrDown;
begin
  inherited Create;
  Parent := nil;
  Xmin := 0;
  Xmax := 0;
  Ymin := 0;
  Ymax := 0;
  Tree := ATree;
  for ns := North to South do
  begin
    for ew := East to West do
    begin
      for ud := Up to Down do
      begin
        Children[ns, ew, ud] := nil;
      end;
    end;
  end;
  FNumPts := 0;

  // Allocate space for the points.
  SetLength(Pts, Tree.MaxPoints);
end;

// Create and initialize the node.
constructor TOtreeNode.Create(const x_min, x_max, y_min, y_max, z_min, z_max : double;
  const ParentNode : TOtreeNode; ATree : TRbwOctTree);
begin
  Create(ATree);
  if not ((x_min <= x_max) and (y_min <= y_max) and (z_min <= z_max)) then
  begin
    raise EOTreeError.Create('Error: attempt to create a TOtreeNode with '
      + 'an invalid range.');
  end;
  Parent := ParentNode;
  Xmin := x_min;
  Xmax := x_max;
  Ymin := y_min;
  Ymax := y_max;
  Zmin := z_min;
  Zmax := z_max;
end;

// Free all children.
destructor TOtreeNode.Destroy;
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
  ud : TUpOrDown;
  Index : integer;
//  ListIndex : integer;
begin
  // If we have no children, free the point memory.
  if (Children[North, West, Up] = nil) then
  begin
    // If required, destroy the data.
    for Index := 0 to NumPts-1 do
    begin
      Pts[Index].Data.Free;
    end;
    SetLength(Pts, 0);
  end
  else
  begin
      // Otherwise free the children.
      for ns := North to South do
        for ew := East to West do
            for ud := Up to Down do
              Children[ns, ew, ud].Free;
      // The Point memory doesn't have to be freed because it was
      // freed when the children were created;
  end;

  inherited Destroy;
end;

procedure TOtreeNode.RemovePoint(X, Y, Z : double; Data : pointer);
var
  xmidLocal, ymidLocal, zmidLocal : double;
  ns            : TNorthOrSouth;
  ew            : TEastOrWest;
  ud : TUpOrDown;
  AnOctNode : TOtreeNode;
  Index, PointIndex : integer;
begin
  // If the point is outside the range of the TOtreeNode, expand the TOtreeNode to encompass it.
  if (X < Xmin) or (X > Xmax) or (Y < Ymin) or (Y > Ymax) then Exit;

  AnOctNode := self;
  // See if the node belongs in a child.
  While (AnOctNode.Children[North, West,Up] <> nil) do

  begin
    xmidLocal := AnOctNode.Xmid;
    ymidLocal := AnOctNode.Ymid;
    zmidLocal := AnOctNode.Zmid;

    // Check that children exist.

    // Add the point to the correct child.
    if (X <= xmidLocal) then
        ew := West
    else
        ew := East;
    if (Y <= ymidLocal) then
        ns := South
    else
        ns := North;
    if (Z <= zmidLocal) then
        ud := Down
    else
        ud := Up;
    AnOctNode := AnOctNode.Children[ns, ew, ud];
  end;

  // find point.
  for Index := 0 to AnOctNode.NumPts -1 do
  begin
    if (AnOctNode.Pts[Index].X = X)
      and (AnOctNode.Pts[Index].Y = Y)
      and (AnOctNode.Pts[Index].Z = Z) then
    begin
      // Remove the data
      AnOctNode.Pts[Index].Data.Remove(Data);
{      if Assigned(Tree.FDestroyProcedure) then
      begin
        Tree.FDestroyProcedure(Tree, Data);
      end;   }
      // if the number of data items associated with the data
      // point has bee reduced to 0, remove the data point.
      if AnOctNode.Pts[Index].Data.Count = 0 then
      begin
        for PointIndex := Index + 1 to AnOctNode.NumPts -1 do
        begin
          AnOctNode.Pts[PointIndex-1] := AnOctNode.Pts[PointIndex];
        end;
        AnOctNode.Pts[Index].Data.Free;
        AnOctNode.Pts[Index].Data := nil;
        AnOctNode.NumPts := AnOctNode.NumPts -1;
      end;
      Exit;
    end;
  end;
end;

procedure TOtreeNode.AddPoint(X, Y, Z : double; Data : pointer);
var
    xmidLocal, ymidLocal, zmidLocal : double;
  i : integer;
    ns            : TNorthOrSouth;
    ew            : TEastOrWest;
  ud : TUpOrDown;
  AnOctNode, Child : TOtreeNode;
  Index : integer;
  APoint : TOPoint;
begin
    // If the point is outside the range of the TOtreeNode, expand the TOtreeNode to encompass it.
    if X < Xmin then ExpandBounds(X, ed3West);
    if X > Xmax then ExpandBounds(X, ed3East);
    if Y < Ymin then ExpandBounds(Y, ed3South);
    if Y > Ymax then ExpandBounds(Y, ed3North);
    if Z < Zmin then ExpandBounds(Z, ed3Down);
    if Z > Zmax then ExpandBounds(Z, ed3Up);

    AnOctNode := self;
    // See if the node belongs in a child.
    While (AnOctNode.NumPts >= Tree.MaxPoints) do

    begin
        xmidLocal := AnOctNode.Xmid;
        ymidLocal := AnOctNode.Ymid;
        zmidLocal := AnOctNode.Zmid;

        // See if we need to create new children.
        if (AnOctNode.Children[North, West, Up] = nil) then
        begin
            // Split the node into eight children.
            AnOctNode.Children[North, West,Up] := TOtreeNode.Create
              (AnOctNode.Xmin, xmidLocal, ymidLocal, AnOctNode.Ymax,
              zmidLocal, AnOctNode.Zmax, AnOctNode, Tree);
            AnOctNode.Children[North, East,Up] := TOtreeNode.Create
              (xmidLocal, AnOctNode.Xmax, ymidLocal, AnOctNode.Ymax,
              zmidLocal, AnOctNode.Zmax, AnOctNode, Tree);
            AnOctNode.Children[South, West,Up] := TOtreeNode.Create
              (AnOctNode.Xmin, xmidLocal, AnOctNode.Ymin, ymidLocal,
              zmidLocal, AnOctNode.Zmax, AnOctNode, Tree);
            AnOctNode.Children[South, East,Up] := TOtreeNode.Create
              (xmidLocal, AnOctNode.Xmax, AnOctNode.Ymin, ymidLocal,
              zmidLocal, AnOctNode.Zmax, AnOctNode, Tree);
            AnOctNode.Children[North, West,Down] := TOtreeNode.Create
              (AnOctNode.Xmin, xmidLocal, ymidLocal, AnOctNode.Ymax,
              AnOctNode.Zmin, zmidLocal, AnOctNode, Tree);
            AnOctNode.Children[North, East,Down] := TOtreeNode.Create
              (xmidLocal, AnOctNode.Xmax, ymidLocal, AnOctNode.Ymax,
              AnOctNode.Zmin, zmidLocal, AnOctNode, Tree);
            AnOctNode.Children[South, West,Down] := TOtreeNode.Create
              (AnOctNode.Xmin, xmidLocal, AnOctNode.Ymin, ymidLocal,
              AnOctNode.Zmin, zmidLocal, AnOctNode, Tree);
            AnOctNode.Children[South, East,Down] := TOtreeNode.Create
              (xmidLocal, AnOctNode.Xmax, AnOctNode.Ymin, ymidLocal,
              AnOctNode.Zmin, zmidLocal, AnOctNode, Tree);

            // Move the old points into the new children.
            for i := 0 to AnOctNode.NumPts-1 do
            begin
              APoint := AnOctNode.Pts[i];
              if (APoint.X <= xmidLocal) then
                  ew := West
              else
                  ew := East;
              if (APoint.Y <= ymidLocal) then
                  ns := South
              else
                  ns := North;
              if (APoint.Z <= zmidLocal) then
                  ud := Down
              else
                  ud := Up;
              Child := AnOctNode.Children[ns, ew, ud];
              Child.Pts[Child.NumPts] := APoint;
              Child.FNumPts := Child.FNumPts + 1;
            end; // End moving points to the new children.

            // Free point memory.
            SetLength(AnOctNode.Pts, 0);

        end; // End creating new children.

        // Add the point to the correct child.
        if (X <= xmidLocal) then
            ew := West
        else
            ew := East;
        if (Y <= ymidLocal) then
            ns := South
        else
            ns := North;
        if (Z <= zmidLocal) then
            ud := Down
        else
            ud := Up;
        AnOctNode := AnOctNode.Children[ns, ew, ud];
    end;

    // Don't add duplicate point. Instead add data to existing point.
    for Index := 0 to AnOctNode.NumPts -1 do
    begin
      if (AnOctNode.Pts[Index].X = X)
        and (AnOctNode.Pts[Index].Y = Y)
        and (AnOctNode.Pts[Index].Z = Z) then
      begin
        AnOctNode.Pts[Index].Data.Add(Data);
        Exit;
      end;
    end;

    // Place the point in this node.
    AnOctNode.Pts[AnOctNode.NumPts].X    := X;
    AnOctNode.Pts[AnOctNode.NumPts].Y    := Y;
    AnOctNode.Pts[AnOctNode.NumPts].Z    := Z;
    AnOctNode.Pts[AnOctNode.NumPts].Data := TList.Create;
    AnOctNode.Pts[AnOctNode.NumPts].Data.Add(Data);
    AnOctNode.NumPts := AnOctNode.NumPts + 1;
end;


// Find the point closest to the given coordinates.
function TOtreeNode.FindClosestPoint(const X, Y, Z : double) : TOPoint;
var
  best_dist2, best_dist : double;
  best_i     : Integer;
  leaf, Aleaf                   : TOtreeNode;
  Siblings : TStack;
begin
  // See which leaf contains the point.
  Siblings := TStack.Create;
  try

    leaf := LocateLeaf(X, Y, Z, Siblings);

    // Find the closest point within the leaf.
    leaf.NearPointInLeaf(X, Y, Z, best_i, best_dist2);

    // If no point was found, check the siblings of the leaf.
    if best_i < 0 then
    begin
      while Siblings.Count > 0 do
      begin
        Aleaf := Siblings.Pop;
        leaf := Aleaf.LocateLeaf(X, Y, Z, Siblings);

        // Find the closest point within the leaf.
        leaf.NearPointInLeaf(X, Y, Z, best_i, best_dist2);
        if best_i >= 0 then
        begin
          Break;
        end;
      end;
    end;

    best_dist := Sqrt(best_dist2);
    // Check nearby leaves for closer points.
    CheckNearbyLeaves(leaf, leaf, X, Y, Z, best_i, best_dist2, best_dist);

    Assert((best_i >= 0) and (best_i < Length(leaf.Pts)));
    result    := leaf.Pts[best_i];
  finally
    Siblings.Free;
  end;
end;

procedure TOtreeNode.FindPoint(var X, Y, Z : double; var Data : pointer);
var
  APoint : TOPoint;
begin
  APoint := FindClosestPoint(X,Y,Z);
  X    := APoint.X;
  Y    := APoint.Y;
  Z    := APoint.Z;
  if (APoint.Data <> nil) and (APoint.Data.Count > 0) then
  begin
    Data := APoint.Data[0];
  end
  else
  begin
    Data := nil;
  end;
end;

// See what leaf contains the point.
function TOtreeNode.LocateLeaf(X, Y, Z : double; Siblings : TStack) : TOtreeNode;
var
    ns, nsi         : TNorthOrSouth;
    ew, ewi         : TEastOrWest;
    ud, udi         : TUpOrDown;
begin
  Result := Self;

  // Search the appropriate child.
  while result.Children[North, West, Up] <> nil do
  begin
    if (X <= result.xmid) then
        ew := West
    else
        ew := East;
    if (Y <= result.ymid) then
        ns := South
    else
        ns := North;
    if (Z <= result.zmid) then
        ud := Down
    else
        ud := Up;

    // if the appropriate child doesn't have any points, we may
    // have to seach the child's siblings so save them.
    for nsi := North to South do
    begin
      for ewi := East to West do
      begin
        for udi := Up to Down do
        begin
          if (nsi <> ns) or (ewi <> ew) or (udi <> ud) then
          begin
            Siblings.Push(result.Children[nsi, ewi, udi]);
          end;
        end;
      end;
    end;
    Result := result.Children[ns, ew, ud];
  end;
end;

// Return the index of the point closest to the given
// coordinates in this leaf.
procedure TOtreeNode.NearPointInLeaf(X, Y, Z : double; var best_i : Integer;
  var best_dist2 : double);
var
  i : Longint;
  dist2 : double;
  dx, dy, dz : double;
begin
  // If there are no points in the node, set best_i to -1 to indicate no
  // points were found.
  best_dist2 := 0;
  best_i := -1;

  if NumPts > 0 then
  begin
    dx := X - Pts[0].X;
    dy := Y - Pts[0].Y;
    dz := Z - Pts[0].Z;
    best_dist2 := dx * dx + dy * dy + dz * dz;
    best_i := 0;

    for i := 1 to NumPts-1 do
    begin
      dx := X - Pts[i].X;
      dy := Y - Pts[i].Y;
      dz := Z - Pts[i].Z;
      dist2 := dx * dx + dy * dy + dz * dz;
      if (best_dist2 > dist2) then
      begin
        best_i := i;
        best_dist2 := dist2;
      end;
    end;
  end;
end;

// Check nearby leaves to see if there is a better point.
procedure TOtreeNode.CheckNearbyLeaves(exclude : TOtreeNode;
  var best_leaf : TOtreeNode; X, Y, Z : double; var best_i : integer;
  var best_dist2, best_dist : double);
var
  xmidLocal, ymidLocal, zmidLocal : double;
  i : Integer;
  dist2 : double;
begin
  // If we are excluding this leaf, do nothing.
  if (exclude = Self) then exit;

  // If this is a leaf node, look for close nodes.
  if (Children[North, West, Up] = nil) then
  begin
    NearPointInLeaf(X, Y, Z, i, dist2);
    // Use the closest point in this leaf if
    // 1. there is at least one point in this leaf (i > -1) and either
    // 2. the closest point in this leaf is closer to the search point than
    //    the point that was already found (best_dist2 > dist2) or
    // 3. no point has been found yet (best_i < 0).
    if (i > -1) and ((best_dist2 > dist2) or (best_i < 0)) then
    begin
      best_dist2 := dist2;
      best_dist := Sqrt(best_dist2);
      best_leaf := Self;
      best_i := i;
    end;
  end
  else
  begin
    // Examine children that lie within best_dist
    // of the point.
    xmidLocal := Xmid;
    ymidLocal := Ymid;
    zmidLocal := Zmid;
    if (best_i < 0) or (X - best_dist <= xmidLocal) then
    begin
      // The West children may be close enough.
      // See if the SouthWest child is close enough.
      if (Y - best_dist <= ymidLocal) or (best_i < 0) Then
      begin
        // See if the SouthWestDown child is close enough.

        if (Z - best_dist <= ZmidLocal) or (best_i < 0) Then
        begin

          Children[South, West, Down].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;

    if (best_i < 0) or (X - best_dist <= xmidLocal) then
    begin
      // The West children may be close enough.
      // See if the SouthWest child is close enough.
      if (Y - best_dist <= ymidLocal) or (best_i < 0) Then
      begin
        // See if the SouthWestUp child is close enough.

        if (Z + best_dist >= ZmidLocal) or (best_i < 0) Then
        begin

          Children[South, West, Up].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;

    if (best_i < 0) or (X - best_dist <= xmidLocal) then
    begin
      // The West children may be close enough.
      // See if the NorthWest child is close enough.
      if (Y + best_dist >= ymidLocal) or (best_i < 0) Then
      begin
        // See if the NorthWestDown child is close enough.
        if (Z - best_dist <= zmidLocal) or (best_i < 0) Then
        begin
          Children[North, West, Down].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;

    if (best_i < 0) or (X - best_dist <= xmidLocal) then
    begin
      // The West children may be close enough.
      // See if the NorthWest child is close enough.
      if (Y + best_dist >= ymidLocal) or (best_i < 0) Then
      begin
        // See if the NorthWestUp child is close enough.
        if (Z + best_dist >= zmidLocal) or (best_i < 0) Then
        begin
          Children[North, West, Up].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;




    if (best_i < 0) or (X + best_dist >= xmidLocal) then
    begin
      // The East children may be close enough.
      // See if the SouthEast child is close enough.
      if (Y - best_dist <= ymidLocal) or (best_i < 0) Then
      begin
        // See if the SouthEastDown child is close enough.
        if (Z - best_dist <= zmidLocal) or (best_i < 0) Then
        begin
          Children[South, East, Down].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;

    if (best_i < 0) or (X + best_dist >= xmidLocal) then
    begin
      // The East children may be close enough.
      // See if the SouthEast child is close enough.
      if (Y - best_dist <= ymidLocal) or (best_i < 0) Then
      begin
        // See if the SouthEastUp child is close enough.
        if (Z + best_dist >= zmidLocal) or (best_i < 0) Then
        begin
          Children[South, East, Up].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;

    if (best_i < 0) or (X + best_dist >= xmidLocal) then
    begin
      // The East children may be close enough.
      // See if the NorthEast child is close enough.
      if (Y + best_dist >= ymidLocal) or (best_i < 0) Then
      begin
        if (Z - best_dist <= zmidLocal) or (best_i < 0) Then
        begin
          Children[North, East, Down].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;

    if (best_i < 0) or (X + best_dist >= xmidLocal) then
    begin
      // The East children may be close enough.
      // See if the NorthEast child is close enough.
      if (Y + best_dist >= ymidLocal) or (best_i < 0) Then
      begin
        if (Z + best_dist >= zmidLocal) or (best_i < 0) Then
        begin
          Children[North, East, Up].CheckNearbyLeaves(
            exclude, best_leaf, X, Y, Z, best_i,
            best_dist2, best_dist);
        end;
      end;
    end;

  end; // End if a leaf ... else check children ...
end;

procedure TOtreeNode.ExpandBounds(XYZ : double;
  ExpandDirection : TExpandDirection3);
begin
  case ExpandDirection of
    ed3North:
      begin
        Ymax := XYZ;
        if Children[North, East,Up] <> nil then
        begin
          Children[North, East,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[North, West,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[North, East,Down].ExpandBounds(XYZ, ExpandDirection);
          Children[North, West,Down].ExpandBounds(XYZ, ExpandDirection);
        end;
      end;
    ed3South:
      begin
        Ymin := XYZ;
        if Children[North, East,Up] <> nil then
        begin
          Children[South, East,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[South, West,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[South, East,Down].ExpandBounds(XYZ, ExpandDirection);
          Children[South, West,Down].ExpandBounds(XYZ, ExpandDirection);
        end;
      end;
    ed3East:
      begin
        Xmax := XYZ;
        if Children[North, East,Up] <> nil then
        begin
          Children[North, East,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[South, East,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[North, East,Down].ExpandBounds(XYZ, ExpandDirection);
          Children[South, East,Down].ExpandBounds(XYZ, ExpandDirection);
        end;
      end;
    ed3West:
      begin
        Xmin := XYZ;
        if Children[North, East,Up] <> nil then
        begin
          Children[North, West,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[South, West,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[North, West,Down].ExpandBounds(XYZ, ExpandDirection);
          Children[South, West,Down].ExpandBounds(XYZ, ExpandDirection);
        end;
      end;
    ed3Up:
      begin
        Zmax := XYZ;
        if Children[North, East,Up] <> nil then
        begin
          Children[North, East,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[South, East,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[North, West,Up].ExpandBounds(XYZ, ExpandDirection);
          Children[South, West,Up].ExpandBounds(XYZ, ExpandDirection);
        end;
      end;
    ed3Down:
      begin
        Zmin := XYZ;
        if Children[North, East,Up] <> nil then
        begin
          Children[North, East,Down].ExpandBounds(XYZ, ExpandDirection);
          Children[South, East,Down].ExpandBounds(XYZ, ExpandDirection);
          Children[North, West,Down].ExpandBounds(XYZ, ExpandDirection);
          Children[South, West,Down].ExpandBounds(XYZ, ExpandDirection);
        end;
      end;
  end;
end;

function TOtreeNode.Xmid : double;
begin
  if Children[North, West, Up] = nil then
  begin
    result := (XMax + XMin) / 2
  end
  else
  begin
    result := Children[North, West, Up].XMax;
  end;
end;

function TOtreeNode.Ymid : double;
begin
  if Children[North, East, Up] = nil then
  begin
    result := (YMax + YMin) / 2
  end
  else
  begin
    result := Children[North, East, Up].YMin;
  end;
end;

function TOtreeNode.Zmid : double;
begin
  if Children[North, East, Up] = nil then
  begin
    result := (ZMax + ZMin) / 2
  end
  else
  begin
    result := Children[North, East, Up].ZMin;
  end;
end;

{ TRbwOctTree }

function TRbwOctTree.GetXMax : double;
begin
  result := FOTreeNode.XMax;
end;

function TRbwOctTree.GetXMin : double;
begin
  result := FOTreeNode.XMin;
end;

function TRbwOctTree.GetYMax : double;
begin
  result := FOTreeNode.YMax;
end;

function TRbwOctTree.GetYMin : double;
begin
  result := FOTreeNode.YMin;
end;

function TRbwOctTree.GetZMax : double;
begin
  result := FOTreeNode.ZMax;
end;

function TRbwOctTree.GetZMin : double;
begin
  result := FOTreeNode.ZMin;
end;


procedure TRbwOctTree.SetXMax(const AValue : double);
begin
  if Count = 0 then
  begin
    FOTreeNode.Xmax := AValue;
  end
  else
  begin
    if (AValue < XMax) then
    begin
      raise EOTreeError.Create('Error: The maximum X value can not be decreased; '
        + 'only increased.');
    end;
    FOTreeNode.ExpandBounds(AValue, ed3West);
  end;
end;

procedure TRbwOctTree.SetXMin(const AValue : double);
begin
  if Count = 0 then
  begin
    FOTreeNode.Xmin := AValue;
  end
  else
  begin
    if (AValue > XMin) then
    begin
      raise EOTreeError.Create('Error: The minimum X value can not be increased; '
        + 'only decreased.');
    end;
    FOTreeNode.ExpandBounds(AValue, ed3East);
  end;
end;

procedure TRbwOctTree.SetYMax(const AValue : double);
begin
  if Count = 0 then
  begin
    FOTreeNode.Ymax := AValue;
  end
  else
  begin
    if (AValue < YMax) then
    begin
      raise EOTreeError.Create('Error: The maximum Y value can not be decreased; '
        + 'only increased.');
    end;
    FOTreeNode.ExpandBounds(AValue, ed3North);
  end;
end;

procedure TRbwOctTree.SetYMin(const AValue : double);
begin
  if Count = 0 then
  begin
    FOTreeNode.Ymin := AValue;
  end
  else
  begin
    if (AValue > YMin) then
    begin
      raise EOTreeError.Create('Error: The minimum Y value can not be increased; '
        + 'only decreased.');
    end;
    FOTreeNode.ExpandBounds(AValue, ed3South);
  end;
end;

procedure TRbwOctTree.SetZMax(const AValue : double);
begin
  if Count = 0 then
  begin
    FOTreeNode.Zmax := AValue;
  end
  else
  begin
    if (AValue < ZMax) then
    begin
      raise EOTreeError.Create('Error: The maximum X value can not be decreased; '
        + 'only increased.');
    end;
    FOTreeNode.ExpandBounds(AValue, ed3Up);
  end;
end;

procedure TRbwOctTree.SetZMin(const AValue : double);
begin
  if Count = 0 then
  begin
    FOTreeNode.Zmin := AValue;
  end
  else
  begin
    if (AValue > ZMin) then
    begin
      raise EOTreeError.Create('Error: The minimum X value can not be increased; '
        + 'only decreased.');
    end;
    FOTreeNode.ExpandBounds(AValue, ed3Down);
  end;
end;

procedure TRbwOctTree.AddPoint(X, Y, Z : double; Data : pointer);
begin
  FOTreeNode.AddPoint(X, Y, Z, Data );
end;

constructor TRbwOctTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMaxPoints := 100;
//  FDestroyProcedure := nil;
  FOTreeNode := TOtreeNode.Create(self);
end;

destructor TRbwOctTree.Destroy;
begin
  FOTreeNode.Free;
  inherited;
end;

function TRbwOctTree.NearestPointsFirstData(X, Y, Z : double) : pointer;
begin
  if FOTreeNode.NumPts > 0 then
  begin
    FirstNearestPoint(X, Y, Z, result);
  end
  else
  begin
    result := nil;
  end;
end;

procedure TRbwOctTree.FirstNearestPoint(var X, Y, Z : double; var Data : pointer);
begin
  if FOTreeNode.NumPts > 0 then
  begin
    FOTreeNode.FindPoint(X, Y, Z, Data);
  end
  else
  begin
    raise EOTreeError.Create('Error: No data points in OctTree.');
  end;
end;

procedure TOtreeNode.SetNumPts(const Value: Integer);
var
  OldNumPoints {, Delta} : Integer;
  ew : TEastOrWest;
  ns : TNorthOrSouth;
  ud : TUpOrDown;
  AChild : TOtreeNode;
  Index : integer;
  PtCount : integer;
begin
  if FNumPts <> Value then
  begin
    OldNumPoints := FNumPts;
    FNumPts := Value;

    if (FNumPts <= Tree.MaxPoints) and (OldNumPoints > Tree.MaxPoints) then
    begin
      SetLength(Pts, Tree.MaxPoints);
      PtCount := 0;
      for ns := North to South do
      begin
        for ew := East to West do
        begin
          for ud := Up to Down do
          begin
            AChild := Children[ns, ew, ud];
            for Index := 0 to AChild.NumPts -1 do
            begin
              Assert(PtCount < Tree.MaxPoints);
              Pts[PtCount] :=  AChild.Pts[Index];
              Inc(PtCount);
            end;
            AChild.FNumPts := 0;
            SetLength(AChild.Pts, 0);
            AChild.Free;
            Children[ns, ew, ud] := nil;
          end;
        end;
      end;
      Assert(PtCount = FNumPts);
    end;

    if Parent <> nil then
    begin
      Parent.ResetCount;
    end;
  end;
end;

function TOtreeNode.GetPoints(Index: Integer): TOPoint;
var
  ew : TEastOrWest;
  ns : TNorthOrSouth;
  ud : TUpOrDown;
  AChild : TOtreeNode;
begin
  if Children[North,East,Up] = nil then
  begin
    result := Pts[Index];
  end
  else
  begin
    for ew := West downto East do
    begin
      for ns := North to South do
      begin
        for ud := Up to Down do
        begin
          AChild := Children[ns,ew,ud];
          if AChild.NumPts > Index then
          begin
            result := AChild.GetPoints(Index);
            Exit;
          end
          else
          begin
            Index := Index - AChild.NumPts;
          end;
        end;
      end;
    end;
  end;
end;

procedure TOtreeNode.FindClosestPointsData(var X, Y, Z: double;
  var Data: TPointerArray);
var
  APoint : TOPoint;
  Index : integer;
begin
  APoint := FindClosestPoint(X,Y,Z);
  X    := APoint.X;
  Y    := APoint.Y;
  Z    := APoint.Z;
  SetLength(Data, APoint.Data.Count);
  for Index := 0 to APoint.Data.Count -1 do
  begin
    Data[Index] := APoint.Data[Index];
  end;
end;

procedure TRbwOctTree.FindClosestPointsData(var X, Y, Z: double;
  var Data: TPointerArray);
begin
  if FOTreeNode.NumPts > 0 then
  begin
    FOTreeNode.FindClosestPointsData(X, Y, Z, Data);
  end
  else
  begin
    raise EOTreeError.Create('Error: No data points in OctTree.');
  end;
end;

function TRbwOctTree.GetCount: integer;
begin
  result := FOTreeNode.NumPts;
end;

procedure TRbwOctTree.RemovePoint(X, Y, Z: double; Data: pointer);
begin
  FOTreeNode.RemovePoint(X, Y, Z, Data );
end;

function TRbwOctTree.GetPoints(Index: Integer): TOctPoint;
var
  OPoint : TOPoint;
  DataIndex : integer;
begin
  if Index < 0 then
  begin
    raise EOTreeError.Create('Invalid point index < 0.');
  end;
  if Index >= Count then
  begin
    raise EOTreeError.Create('Invalid point index >= Count.');
  end;
  OPoint := FOTreeNode.Points[Index];
  result.X := OPoint.X;
  result.Y := OPoint.Y;
  result.Z := OPoint.Z;
  SetLength(result.Data, OPoint.Data.Count);
  for DataIndex := 0 to OPoint.Data.Count -1 do
  begin
    result.Data[DataIndex] := OPoint.Data[DataIndex];
  end;
end;

procedure TOtreeNode.ResetCount;
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
  ud : TUpOrDown;
  Count : integer;
begin
  if Children[North, West, Up] <> nil then
  begin
    Count := 0;
    for ns := North to South do
    begin
      for ew := East to West do
      begin
        for ud := Up to Down do
        begin
          Count := Count + Children[ns, ew, ud].NumPts;
        end;
      end;
    end;
    NumPts := Count;
  end;
end;

{procedure TRbwOctTree.SetDestroyProcedure(
  const Value: TDestroyDataProcedure);
begin
  FDestroyProcedure := Value;
end; }

procedure TOtreeNode.FindNearestPoints(const CenterX, CenterY, CenterZ : double;
  const Count: Integer; const List: TObjectList);
var
  Radius : double;
  Node : TSelectNode;
  distance : double;
  APoint : TOPoint;
  dx, dy, dz : double;
  ALeaf : TOtreeNode;
  Siblings : TStack;
  Procedure GetPoints;
  var
    Index, InnerIndex : integer;
  begin
    for Index := 0 to ALeaf.NumPts -1 do
    begin
      APoint := ALeaf.Pts[Index];
      dx := APoint.X - CenterX;
      dy := APoint.Y - CenterY;
      dz := APoint.Z - CenterZ;
      distance := Sqrt(Sqr(dx) + Sqr(dy) + Sqr(dz));
      if (List.Count < Count) or (distance <= Radius) then
      begin
        Node := TSelectNode.Create;
        Node.Distance := distance;
        Node.Point := APoint;
        List.Add(Node);
        if List.Count mod Count = 0 then
        begin
          List.Sort(SelectNodeCompare);
          if List.Count >= Count then
          begin
            Node := List[Count-1] as TSelectNode;
            Radius := Node.Distance;
          end;
          for InnerIndex := Count to List.Count - 1 do
          begin
            Node := List[InnerIndex] as TSelectNode;
            if Node.Distance > Radius then
            begin
              List.Count := InnerIndex;
              break;
            end;
          end;
        end;
      end;
    end;
  end;
begin
  Siblings := TStack.Create;
  try
    ALeaf := LocateLeaf(CenterX, CenterY, CenterZ, Siblings);
    radius := 0;
    GetPoints;
    While Siblings.Count > 0 do
    begin
      ALeaf := Siblings.Pop;

      if (List.Count < Count) or
        ((CenterX + radius >= ALeaf.Xmin) and
        (CenterX - radius <= ALeaf.Xmax) and
        (CenterY + radius >= ALeaf.Ymin) and
        (CenterY - radius <= ALeaf.Ymax) and
        (CenterZ + radius >= ALeaf.Zmin) and
        (CenterZ - radius <= ALeaf.Zmax)) then
      begin
        ALeaf := ALeaf.LocateLeaf(CenterX, CenterY, CenterZ, Siblings);
        GetPoints;
      end;
    end;
  finally
    Siblings.Free;
  end;
end;

procedure TOtreeNode.FindPointsInSphere(const CenterX, CenterY, CenterZ,
  Radius, RadiusSquared: double; const List: TList);
var
  AChild : TOtreeNode;
  xmidLocal, ymidLocal, zmidLocal : double;
  Index : integer;
  dx, dy, dz : double;
  APoint : TOPoint;
  WholeBlockInside : boolean;
begin
  if Children[North, West, Up] = nil then
  begin
    // see if it is worth trying to determine if all the points
    // are inside the circle;
    WholeBlockInside := NumPts > 8;

    // Test each corner to see whether it is inside the circle
    if WholeBlockInside then
    begin
      dx := XMax - CenterX;
      dy := YMax - CenterY;
      dZ := ZMax - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMax - CenterX;
      dy := YMin - CenterY;
      dZ := ZMax - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMin - CenterX;
      dy := YMax - CenterY;
      dZ := ZMax - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMin - CenterX;
      dy := YMin - CenterY;
      dZ := ZMax - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMax - CenterX;
      dy := YMax - CenterY;
      dZ := ZMin - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMax - CenterX;
      dy := YMin - CenterY;
      dZ := ZMin - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMin - CenterX;
      dy := YMax - CenterY;
      dZ := ZMin - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;

    if WholeBlockInside then
    begin
      dx := XMin - CenterX;
      dy := YMin - CenterY;
      dZ := ZMin - CenterZ;
      if Sqr(dx) + Sqr(dy) + Sqr(dz) > RadiusSquared then
      begin
        WholeBlockInside := False;
      end;
    end;


    // If the entire block is inside the circle, add all the points.
    // Otherwise, test each point to see if it is inside the circle and
    // add those points.
    if WholeBlockInside then
    begin
      for Index := 0 to NumPts -1 do
      begin
        List.Add(Addr(Pts[Index]));
      end;
    end
    else
    begin
      for Index := 0 to NumPts -1 do
      begin
        APoint := Pts[Index];
        dx := APoint.X - CenterX;
        dy := APoint.Y - CenterY;
        dz := APoint.Z - CenterZ;
        if Sqr(dx) + Sqr(dy) + Sqr(dz) <= RadiusSquared then
        begin
          List.Add(Addr(Pts[Index]));
        end;
      end;
    end;
  end
  else
  begin
    xmidLocal := Xmid;
    ymidLocal := Ymid;
    zmidLocal := Zmid;
    if (CenterX - Radius <= xmidLocal) then
    begin
      // West Children are candidates.
      if CenterY - Radius <= ymidLocal then
      begin
        //South Children are candidates.
        if CenterZ - Radius <= zmidLocal then
        begin
          // Down Children are candidates
          AChild := Children[South, West, Down];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;

        if CenterZ + Radius >= zmidLocal then
        begin
          // Up Children are candidates
          AChild := Children[South, West, Up];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;
      end;

      if CenterY + Radius >= ymidLocal then
      begin
        //North Children are candidates.
        if CenterZ - Radius <= zmidLocal then
        begin
          // Down Children are candidates
          AChild := Children[North, West, Down];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;
        if CenterZ + Radius >= zmidLocal then
        begin
          // Up Children are candidates
          AChild := Children[North, West, Up];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;
      end;
    end;

    if (CenterX + Radius >= xmidLocal) then
    begin
      // East Children are candidates.
      if CenterY - Radius <= ymidLocal then
      begin
        //South Children are candidates.
        if CenterZ - Radius <= zmidLocal then
        begin
          // Down Children are candidates
          AChild := Children[South, East, Down];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;
        if CenterZ + Radius >= zmidLocal then
        begin
          // Up Children are candidates
          AChild := Children[South, East, Up];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;
      end;

      if CenterY + Radius >= ymidLocal then
      begin
        //North Children are candidates.
        if CenterZ - Radius <= zmidLocal then
        begin
          // Down Children are candidates
          AChild := Children[North, East, Down];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;
        if CenterZ + Radius >= zmidLocal then
        begin
          // Up Children are candidates
          AChild := Children[North, East, Up];
          AChild.FindPointsInSphere(CenterX, CenterY, CenterZ, Radius,
            RadiusSquared, List);
        end;
      end;
    end
  end;
end;

procedure TOtreeNode.FindPointsInBlock(const Block: T3DBlock;
  const List: TList);
var
  AChild : TOtreeNode;
  xmidLocal, ymidLocal, zmidLocal : double;
  Index : integer;
//  dx, dy : double;
  APoint : TOPoint;
begin
  if Children[North, West, Up] = nil then
  begin
    if (Block.XMin <= XMin) and (Block.XMax >= XMax)
      and (Block.YMin <= YMin) and (Block.YMax >= YMax)
      and (Block.ZMin <= ZMin) and (Block.ZMax >= ZMax) then
    begin
      for Index := 0 to NumPts -1 do
      begin
        List.Add(Addr(Pts[Index]));
      end;
    end
    else
    begin
      for Index := 0 to NumPts -1 do
      begin
        APoint := Pts[Index];
        if (Block.XMin <= APoint.X) and (Block.XMax >= APoint.X)
          and (Block.YMin <= APoint.Y) and (Block.YMax >= APoint.Y)
          and (Block.ZMin <= APoint.Z) and (Block.ZMax >= APoint.Z) then
        begin
          List.Add(Addr(Pts[Index]));
        end;
      end;
    end;
  end
  else
  begin
    xmidLocal := Xmid;
    ymidLocal := Ymid;
    zmidLocal := Zmid;
    if (Block.XMin <= xmidLocal) then
    begin
      // West Children are candidates.
      if Block.YMin <= ymidLocal then
      begin
        //South Children are candidates.
        if Block.ZMin <= zmidLocal then
        begin
          //Down Children are candidates
          AChild := Children[South, West, Down];
          AChild.FindPointsInBlock(Block, List);
        end;
        if Block.ZMax >= zmidLocal then
        begin
          //Up Children are candidates
          AChild := Children[South, West, Up];
          AChild.FindPointsInBlock(Block, List);
        end;

      end;

      if Block.YMax >= ymidLocal then
      begin
        //North Children are candidates.
        if Block.ZMin <= zmidLocal then
        begin
          //Down Children are candidates
          AChild := Children[North, West, Down];
          AChild.FindPointsInBlock(Block, List);
        end;
        if Block.ZMax >= zmidLocal then
        begin
          //Up Children are candidates
          AChild := Children[North, West, Up];
          AChild.FindPointsInBlock(Block, List);
        end;
      end;
    end;

    if (Block.XMax >= xmidLocal) then
    begin
      // East Children are candidates.
      if Block.YMin <= ymidLocal then
      begin
        //South Children are candidates.
        if Block.ZMin <= zmidLocal then
        begin
          //Down Children are candidates
          AChild := Children[South, East, Down];
          AChild.FindPointsInBlock(Block, List);
        end;
        if Block.ZMax >= zmidLocal then
        begin
          //Up Children are candidates
          AChild := Children[South, East, Up];
          AChild.FindPointsInBlock(Block, List);
        end;
      end;

      if Block.YMax >= ymidLocal then
      begin
        //North Children are candidates.
        if Block.ZMin <= zmidLocal then
        begin
          //Down Children are candidates
          AChild := Children[North, East, Down];
          AChild.FindPointsInBlock(Block, List);
        end;
        if Block.ZMax >= zmidLocal then
        begin
          //Up Children are candidates
          AChild := Children[North, East, Up];
          AChild.FindPointsInBlock(Block, List);
        end;
      end;
    end
  end;
end;

procedure TRbwOctTree.FindPointsInBlock(const Block: T3DBlock;
  var Points: TOctPointInRegionArray);
var
  List : TList;
  PointIndex, DataIndex : integer;
  APoint : TOPoint;
  PointAddress : POPoint;
begin
  List := TList.Create;
  try
    List.Capacity := Count;
    FOTreeNode.FindPointsInBlock(Block, List);
    SetLength(Points, List.Count);
    for PointIndex := 0 to List.Count -1 do
    begin
      PointAddress := List[PointIndex];
      APoint := PointAddress^;
      Points[PointIndex].X := APoint.X;
      Points[PointIndex].Y := APoint.Y;
      Points[PointIndex].Z := APoint.Z;
      SetLength(Points[PointIndex].Data, APoint.Data.Count);
      for DataIndex := 0 to APoint.Data.Count -1 do
      begin
        Points[PointIndex].Data[DataIndex] := APoint.Data[DataIndex];
      end;
    end;
  finally
    List.Free;
  end;
end;



procedure TRbwOctTree.FindPointsInSphere(const CenterX, CenterY, CenterZ,
  Radius: double; var Points: TOctPointInRegionArray);
var
  List : TList;
  PointIndex, DataIndex : integer;
  APoint : TOPoint;
  PointAddress : POPoint;
begin
  List := TList.Create;
  try
    FOTreeNode.FindPointsInSphere(CenterX, CenterY,  CenterZ, Radius,
      Sqr(Radius), List);
    SetLength(Points, List.Count);
    for PointIndex := 0 to List.Count -1 do
    begin
      PointAddress := List[PointIndex];
      APoint := PointAddress^;
      Points[PointIndex].X := APoint.X;
      Points[PointIndex].Y := APoint.Y;
      Points[PointIndex].Z := APoint.Z;
      SetLength(Points[PointIndex].Data, APoint.Data.Count);
      for DataIndex := 0 to APoint.Data.Count -1 do
      begin
        Points[PointIndex].Data[DataIndex] := APoint.Data[DataIndex];
      end;
    end;
  finally
    List.Free;
  end;
end;

procedure TRbwOctTree.FindNearestPoints(const CenterX, CenterY, CenterZ: double;
  const Count: Integer; var Points: TOctPointArray);
var
  List : TObjectList;
  PointIndex, DataIndex : integer;
  APoint : TOPoint;
  Node : TSelectNode;
  Radius  : double;
begin
  List := TObjectList.Create;
  try
    FOTreeNode.FindNearestPoints(CenterX, CenterY, CenterZ, Count, List);

    List.Sort(SelectNodeCompare);
    if List.Count > Count then
    begin
      Node := List[Count-1] as TSelectNode;
      Radius := Node.Distance;

      for PointIndex := Count to List.Count - 1 do
      begin
        Node := List[PointIndex] as TSelectNode;
        if Node.Distance > Radius then
        begin
          List.Count := PointIndex;
          break;
        end;
      end;
    end;


    SetLength(Points, List.Count);
    for PointIndex := 0 to List.Count -1 do
    begin
      Node := List[PointIndex] as TSelectNode;
      APoint := Node.Point;
      Points[PointIndex].X := APoint.X;
      Points[PointIndex].Y := APoint.Y;
      Points[PointIndex].Z := APoint.Z;
      Points[PointIndex].Distance := Node.Distance;
      SetLength(Points[PointIndex].Data, APoint.Data.Count);
      for DataIndex := 0 to APoint.Data.Count -1 do
      begin
        Points[PointIndex].Data[DataIndex] := APoint.Data[DataIndex];
      end;
    end;

  finally
    List.Free;
  end;
end;

procedure TRbwOctTree.SetMaxPoints(const Value: integer);
begin
  if Count = 0 then
  begin
    if Value <= 0 then
    begin
      raise EOTreeError.Create('The maximum number of points in an Octtree '
      + 'must be greater than zero.');
    end;
    FMaxPoints := Value;
    SetLength(FOTreeNode.Pts, FMaxPoints);
  end
  else
  begin
    raise EOTreeError.Create('You must set the number of points before adding '
      + 'any points to a Octtree.');
  end;

end;

procedure Register;
begin
  RegisterComponents('RBW', [TRbwOctTree]);
end;

procedure TOtreeNode.Clear;
var
  ns : TNorthOrSouth;
  ew : TEastOrWest;
  ud : TUpOrDown;
  Index : integer;
begin
  if Children[North, East, Up] = nil then
  begin
    for Index := 0 to NumPts-1 do
    begin
      Pts[Index].Data.Free;
      Pts[Index].Data := nil;
    end;
    FNumPts := 0;
  end
  else
  begin
    for ns := North to South do
    begin
      for ew := East to West do
      begin
        for ud := Up to Down do
        begin
          Children[ns,ew,ud].Clear;
          Children[ns,ew,ud].Free;
          Children[ns,ew,ud] := nil;
        end;
      end;
    end;
  end;
end;

procedure TRbwOctTree.Clear;
begin
  FOTreeNode.Clear;
  FOTreeNode.FNumPts := 0;
  SetLength(FOTreeNode.Pts, MaxPoints);
end;

end.
